import ftplib
import asyncio
from io import BytesIO
from saviialib.libs.ftp_client.ftp_client_contract import (
    FTPClientContract,
)
from saviialib.libs.ftp_client.types.ftp_client_types import (
    FtpClientInitArgs,
    FtpListFilesArgs,
    FtpReadFileArgs,
)


class FtplibClient(FTPClientContract):
    def __init__(self, args: FtpClientInitArgs) -> None:
        self.host = args.config.ftp_host
        self.port = args.config.ftp_port
        self.password = args.config.ftp_password
        self.user = args.config.ftp_user
        self.client = ftplib.FTP(host=self.host, user=self.user, passwd=self.password)

    async def _async_start(self) -> None:
        try:
            await asyncio.to_thread(self.client.login, self.user, self.password)
        except OSError:
            raise ConnectionRefusedError(
                f"{self.host}:{self.port} isn't active. "
                "Please ensure the server is running and accessible."
            )
        except Exception as error:
            raise ConnectionError(
                f"General connection for {self.host}:{self.port}.", error.__str__()
            )

    async def list_files(self, args: FtpListFilesArgs) -> list[str]:
        try:
            EXCLUDED_NAMES = [".", ".."]
            await self._async_start()
            await asyncio.to_thread(self.client.cwd, args.path)
            filenames = await asyncio.to_thread(self.client.nlst, args.path)
            return [
                filename for filename in filenames if filename not in EXCLUDED_NAMES
            ]
        except Exception as error:
            raise ConnectionAbortedError(error)

    async def read_file(self, args: FtpReadFileArgs) -> bytes:
        await self._async_start()
        try:
            file_content = BytesIO()
            await asyncio.to_thread(
                self.client.retrbinary, "RETR " + args.file_path, file_content.write
            )
            await asyncio.to_thread(file_content.seek, 0)
            file_bytes = await asyncio.to_thread(file_content.read)
            return file_bytes
        except Exception as error:
            raise FileNotFoundError(f"File not found: {args.file_path}") from error



import asyncio
from time import time
from saviialib.libs.zero_dependency.utils.datetime_utils import today, datetime_to_str
from logging import Logger
from saviialib.general_types.error_types.api.epii_api_error_types import (
    BackupEmptyError,
    BackupSourcePathError,
    BackupUploadError,
)
from saviialib.general_types.error_types.common import (
    SharepointClientError,
)
from saviialib.libs.directory_client import DirectoryClient, DirectoryClientArgs
from saviialib.libs.files_client import (
    FilesClient,
    FilesClientInitArgs,
    ReadArgs,
    WriteArgs,
)
from saviialib.libs.sharepoint_client import (
    SharepointClient,
    SharepointClientInitArgs,
    SpUploadFileArgs,
    SpCreateFolderArgs,
)
from saviialib.services.epii.utils.upload_backup_to_sharepoint_utils import (
    calculate_percentage_uploaded,
    count_files_in_directory,
    extract_error_message,
    parse_execute_response,
    show_upload_result,
)

from .types.upload_backup_to_sharepoint_types import (
    UploadBackupToSharepointUseCaseInput,
)


class UploadBackupToSharepointUsecase:
    def __init__(self, input: UploadBackupToSharepointUseCaseInput):
        self.sharepoint_config = input.sharepoint_config
        self.local_backup_source_path = input.local_backup_source_path
        self.sharepoint_destination_path = input.sharepoint_destination_path

        self.files_client = self._initialize_files_client()
        self.dir_client = self._initialize_directory_client()
        self.log_history = []
        self.grouped_files_by_folder = None
        self.total_files = None
        self.logger: Logger = input.logger
        self.sharepoint_client = self._initalize_sharepoint_client()

    def _initalize_sharepoint_client(self):
        return SharepointClient(
            SharepointClientInitArgs(
                self.sharepoint_config, client_name="sharepoint_rest_api"
            )
        )

    def _initialize_directory_client(self):
        return DirectoryClient(DirectoryClientArgs(client_name="os_client"))

    def _initialize_files_client(self):
        return FilesClient(FilesClientInitArgs(client_name="aiofiles_client"))

    async def _initialize_backup_base_folder(self):
        local_backup_name = (
            f"/local-backup-{datetime_to_str(today(), date_format='%m-%d-%Y')}"
        )
        local_backup_destination_path = (
            self.sharepoint_destination_path + local_backup_name
        )
        async with self.sharepoint_client:
            await self.sharepoint_client.create_folder(
                SpCreateFolderArgs(folder_relative_url=local_backup_destination_path)
            )
        self.sharepoint_destination_path = local_backup_destination_path
        base_folder_message = (
            "[local_backup_lib] Creating base folder" + local_backup_name
        )
        self.logger.info(base_folder_message)
        self.log_history.append(base_folder_message)

    async def _validate_backup_structure(self):
        # Initialize the backup folder
        await self._initialize_backup_base_folder()

        # Check if the local path exists in the main directory
        if not await self.dir_client.path_exists(self.local_backup_source_path):
            raise BackupSourcePathError(
                reason=f"'{self.local_backup_source_path}' doesn't exist."
            )

        local_directories = (
            list(self.grouped_files_by_folder.keys())
            if self.grouped_files_by_folder
            else []
        )
        async with self.sharepoint_client:  # type: ignore
            for local_dir in local_directories:
                create_message = (
                    f"[local_backup_lib] Creating a new directory '{local_dir}'."
                )

                self.log_history.append(create_message)
                await self.sharepoint_client.create_folder(
                    SpCreateFolderArgs(
                        folder_relative_url=self.sharepoint_destination_path
                        + "/"
                        + local_dir
                    )
                )
                self.logger.info("[local_backup_lib] %s", create_message)
                self.log_history.append(f"[local_backup_lib] {create_message}")

        # Check if the current folder only have files and each folder exist in Microsoft Sharepoint.
        if self.total_files == 0:
            no_files_message = (
                f"[local_backup_lib] {self.local_backup_source_path} has no files ⚠️"
            )
            self.log_history.append(no_files_message)
            self.logger.debug(no_files_message)
            raise BackupEmptyError

    async def _group_files_by_folder(self) -> dict[str, list[str]]:
        """Groups files by their parent folder."""
        backup_folder_exists = await self.dir_client.path_exists(
            self.local_backup_source_path
        )

        if not backup_folder_exists:
            return {}
        folder_names = await self.dir_client.listdir(self.local_backup_source_path)
        grouped = {}
        for folder_name in folder_names:
            is_folder = await self.dir_client.isdir(
                self.dir_client.join_paths(self.local_backup_source_path, folder_name)
            )
            if not is_folder:
                continue
            grouped[folder_name] = await self.dir_client.listdir(
                self.dir_client.join_paths(self.local_backup_source_path, folder_name)
            )
        return grouped

    async def _save_log_history(self) -> None:
        await self.files_client.write(
            WriteArgs(
                file_name="BACKUP_LOG_HISTORY.log",
                file_content="\n".join(self.log_history),
                mode="w",
            )
        )

    async def _generate_tasks(self) -> list:
        tasks = []
        for folder_name in self.grouped_files_by_folder:  # type: ignore
            count_files_in_dir = await count_files_in_directory(
                self.local_backup_source_path, folder_name
            )
            if count_files_in_dir == 0:
                empty_folder_message = (
                    f"[local_backup_lib] The folder '{folder_name}' is empty ⚠️"
                )
                self.logger.debug(empty_folder_message)
                self.log_history.append(empty_folder_message)
                continue
            extracting_files_message = (
                "[local_backup_lib]"
                + f" Extracting files from '{folder_name} ".center(15, "*")
            )
            self.log_history.append(extracting_files_message)
            self.logger.debug(extracting_files_message)
            for file_name in self.grouped_files_by_folder[folder_name]:  # type: ignore
                tasks.append(self._upload_and_log_progress_task(folder_name, file_name))

        return tasks

    async def export_file_to_sharepoint(
        self, folder_name: str, file_name: str, file_content: bytes
    ) -> tuple[bool, str]:
        """Uploads a file to the specified folder in SharePoint."""
        uploaded = None
        error_message = ""

        try:
            sharepoint_client = SharepointClient(
                SharepointClientInitArgs(
                    self.sharepoint_config, client_name="sharepoint_rest_api"
                )
            )
        except ConnectionError as error:
            raise SharepointClientError(error)

        async with sharepoint_client:
            try:
                folder_url = f"{self.sharepoint_destination_path}/{folder_name}"
                args = SpUploadFileArgs(
                    folder_relative_url=folder_url,
                    file_content=file_content,
                    file_name=file_name,
                )
                await sharepoint_client.upload_file(args)
                uploaded = True
            except ConnectionError as error:
                error_message = str(error)
                uploaded = False

        return uploaded, error_message

    async def _upload_and_log_progress_task(self, folder_name, file_name) -> dict:
        """Task for uploads a file and logs progress."""
        uploading_message = (
            f"[local_backup_lib] Uploading file '{file_name}' from '{folder_name}' "
        )
        self.log_history.append(uploading_message)
        self.logger.debug(uploading_message)
        file_path = self.dir_client.join_paths(
            self.local_backup_source_path, folder_name, file_name
        )
        file_content = await self.files_client.read(ReadArgs(file_path, mode="rb"))
        uploaded, error_message = await self.export_file_to_sharepoint(
            folder_name,
            file_name,
            file_content,  # type: ignore
        )
        result_message = show_upload_result(uploaded, file_name)
        self.logger.debug(result_message)
        self.log_history.append(result_message)
        return {
            "parent_folder": folder_name,
            "file_name": file_name,
            "uploaded": uploaded,
            "error_message": error_message,
        }

    async def retry_upload_failed_files(self, results) -> None:
        failed_files = [item for item in results if not item["uploaded"]]
        tasks = []
        retry_message = f"[local_backup_lib] Retrying upload for {len(failed_files)} failed files... 🚨"
        self.log_history.append(retry_message)
        self.logger.debug(retry_message)
        for file in failed_files:
            tasks.append(
                self._upload_and_log_progress_task(
                    file["parent_folder"], file["file_name"]
                )
            )
        results = await asyncio.gather(*tasks, return_exceptions=True)
        success = calculate_percentage_uploaded(results, self.total_files)  # type: ignore
        if success < 100.0:
            raise BackupUploadError(
                reason=extract_error_message(self.logger, results, success)  # type: ignore
            )
        else:
            successful_upload_retry = (
                "[local_backup_lib] All files uploaded successfully after retry."
            )
            self.log_history.append(successful_upload_retry)
            self.logger.debug(successful_upload_retry)
            await self._save_log_history()
            return parse_execute_response(results)  # type: ignore

    async def execute(self):
        """Exports all files from the local backup folder to SharePoint cloud."""
        start_time = time()
        self.grouped_files_by_folder = await self._group_files_by_folder()
        self.total_files = sum(
            len(files) for files in self.grouped_files_by_folder.values()
        )

        # Check if the current folder only have files and each folder exist in Microsoft Sharepoint.
        await self._validate_backup_structure()

        # Create task for each file stored in the the local backup folder.
        tasks = await self._generate_tasks()

        # Execution of multiple asynchronous tasks for files migration.
        results = await asyncio.gather(*tasks, return_exceptions=True)
        success = calculate_percentage_uploaded(results, self.total_files)  # type: ignore

        if success < 100.0:
            await self.retry_upload_failed_files(results)
        else:
            end_time = time()
            backup_time = end_time - start_time
            successful_backup_message = (
                f"[local_backup_lib] Migration time: {backup_time:.2f} seconds ✨"
            )
            self.log_history.append(successful_backup_message)

            finished_backup_message = (
                "[local_backup_lib] All the files were uploaded successfully 🎉"
            )
            self.log_history.append(finished_backup_message)

            await self._save_log_history()
            return parse_execute_response(results)  # type: ignore



from dataclasses import dataclass, field
from typing import Dict

from saviialib.general_types.api.epii_api_types import EpiiSharepointBackupConfig


@dataclass
class UploadBackupToSharepointControllerInput:
    config: EpiiSharepointBackupConfig
    local_backup_source_path: str
    sharepoint_destination_path: str


@dataclass
class UploadBackupToSharepointControllerOutput:
    message: str
    status: int
    metadata: Dict[str, str] = field(default_factory=dict)



from .thies_bp import THIESDayData
from typing import List
from logging import Logger
from asyncio import to_thread
from saviialib.libs.directory_client import DirectoryClient
from saviialib.libs.zero_dependency.utils.datetime_utils import datetime_to_str, today
from saviialib.libs.files_client import FilesClient, FilesClientInitArgs, WriteArgs


AVG_COLUMNS = {
    "Date": "date",
    "Time": "time",
    "AirTemperature": "air_temperature",
    "Radiation": "radiation",
    "CO2": "carbon_dioxide",
    "Precipitation": "precipitation",
    "WS": "wind_velocity",
    "WD": "wind_direction",
    "Humidity": "humidity",
}

EXT_COLUMNS = {
    "Date": "date",
    "Time": "time",
    "AirTemperature MIN": "air_temperature",
    "AirTemperature MAX": "air_temperature",
    "Radiation MIN": "radiation",
    "Radiation MAX": "radiation",
    "CO2 MIN": "carbon_dioxide",
    "CO2 MAX": "carbon_dioxide",
    "WS MIN": "wind_velocity",
    "WS MAX gust": "wind_velocity",
    "WD MIN": "wind_direction",
    "WD MAX gust": "wind_direction",
    "Humidity MIN": "humidity",
    "Humidity MAX": "humidity",
}

AGG_DICT = {
    "AirTemperature": "mean",
    "AirTemperature MIN": "mean",
    "AirTemperature MAX": "mean",
    "Precipitation": "sum",
    "Humidity": "mean",
    "Humidity MIN": "mean",
    "Humidity MAX": "mean",
    "Radiation": "sum",
    "Radiation MIN": "sum",
    "Radiation MAX": "sum",
    "CO2": "sum",
    "CO2 MIN": "sum",
    "CO2 MAX": "sum",
    "WS": "mean",
    "WS MIN": "mean",
    "WS MAX gust": "mean",
    "WD": "mean",
    "WD MIN": "mean",
    "WD MAX gust": "mean",
}

UNITS = {
    "AirTemperature": "°C",
    "Precipitation": "mm",
    "Humidity": "%",
    "Radiation": "W/m2",
    "CO2": "ppm",
    "WS": "m/s",
    "WD": "°",
}


async def create_thies_daily_statistics_file(
    os_client: DirectoryClient, logger: Logger, daily_files: List[str]
) -> None:
    logger.debug("[thies_synchronization_lib] Creating Daily Statistics ...")
    csv_client = FilesClient(FilesClientInitArgs(client_name="csv_client"))
    filename = datetime_to_str(today(), date_format="%Y%m%d") + ".BIN"
    path_bin_av = os_client.join_paths("thies-daily-files", "ARCH_AV1", filename)
    path_ini_av = os_client.join_paths("thies-daily-files", "ARCH_AV1", "DESCFILE.INI")
    path_bin_ex = os_client.join_paths("thies-daily-files", "ARCH_EX1", filename)
    path_ini_ex = os_client.join_paths("thies-daily-files", "ARCH_EX1", "DESCFILE.INI")

    ext_df = THIESDayData("ex")
    await to_thread(ext_df.read_binfile, path_bin_ex, path_ini_ex)

    avg_df = THIESDayData("av")
    await to_thread(avg_df.read_binfile, path_bin_av, path_ini_av)

    ext_df = ext_df.dataDF[EXT_COLUMNS.keys()]
    avg_df = avg_df.dataDF[AVG_COLUMNS.keys()]

    # Merge both dataframes
    df = avg_df.merge(ext_df, on=["Date", "Time"], how="outer")
    # Set the date as dd.mm.yyyy format.
    df["Date"] = df["Date"].str.replace(
        r"(\d{4})/(\d{2})/(\d{2})", r"\3.\2.\1", regex=True
    )
    df["Hour"] = df["Time"].str[:2]

    # Group by hour.
    hourly_agg = df.groupby(["Date", "Hour"]).agg(AGG_DICT).reset_index()

    rows = []
    # For each attribute in avg_columns (except Date, Time)
    for col, col_id in AVG_COLUMNS.items():
        if col in ["Date", "Time"]:
            continue
        # Determine the corresponding min/max columns if they exist
        min_col = f"{col} MIN"
        max_col = f"{col} MAX"
        mean_col = col
        if col in ["WS", "WD"]:
            max_col += " gust"

        unit = UNITS.get(col, "")

        for idx, row in hourly_agg.iterrows():
            statistic_id = f"sensor.saviia_epii_{col_id}"
            start = f"{row['Date']} {row['Hour']}:00"
            mean = row[mean_col] if mean_col in row else 0
            min_val = row[min_col] if min_col in row else 0
            max_val = row[max_col] if max_col in row else 0

            # If no min/max for this attribute is 0
            if min_col not in row:
                min_val = 0
            if max_col not in row:
                max_val = 0

            if col in ["WD"]:  # Avoid error
                rows.append(
                    {
                        "statistic_id": statistic_id,
                        "unit": unit,
                        "start": start,
                        "min": mean,
                        "max": mean,
                        "mean": mean,
                    }
                )
            else:
                rows.append(
                    {
                        "statistic_id": statistic_id,
                        "unit": unit,
                        "start": start,
                        "min": min_val,
                        "max": max_val,
                        "mean": mean,
                    }
                )

    logger.debug("[thies_synchronization_lib] Saving file in /config folder ...")
    logger.debug(rows[0].keys())
    await csv_client.write(
        WriteArgs(file_name="thies_daily_statistics.tsv", file_content=rows, mode="w")
    )
    logger.debug(
        "[thies_synchronization_lib] thies_daily_statistics.tsv created successfully!"
    )



from saviialib.general_types.error_types.api.epii_api_error_types import (
    SharePointFetchingError,
    SharePointDirectoryError,
    SharePointUploadError,
    ThiesConnectionError,
    ThiesFetchingError,
)
from saviialib.general_types.error_types.common import (
    EmptyDataError,
    FtpClientError,
    SharepointClientError,
)
from saviialib.libs.ftp_client import (
    FTPClient,
    FtpClientInitArgs,
    FtpListFilesArgs,
    FtpReadFileArgs,
)
from saviialib.libs.sharepoint_client import (
    SharepointClient,
    SharepointClientInitArgs,
    SpListFilesArgs,
    SpListFoldersArgs,
    SpUploadFileArgs,
)
from saviialib.libs.directory_client import DirectoryClient, DirectoryClientArgs

from saviialib.libs.files_client import FilesClient, FilesClientInitArgs, WriteArgs
from saviialib.services.epii.use_cases.types import (
    FtpClientConfig,
    SharepointConfig,
    UpdateThiesDataUseCaseInput,
)
from saviialib.services.epii.utils import (
    parse_execute_response,
)
from saviialib.libs.zero_dependency.utils.datetime_utils import today, datetime_to_str
from .components.create_thies_statistics_file import create_thies_daily_statistics_file


class UpdateThiesDataUseCase:
    def __init__(self, input: UpdateThiesDataUseCaseInput):
        self.sharepoint_client = self._initialize_sharepoint_client(
            input.sharepoint_config
        )
        self.logger = input.logger
        self.thies_ftp_client = self._initialize_thies_ftp_client(input.ftp_config)
        self.sharepoint_folders_path = input.sharepoint_folders_path
        self.ftp_server_folders_path = input.ftp_server_folders_path
        self.sharepoint_base_url = f"/sites/{self.sharepoint_client.site_name}"
        self.uploading = set()
        self.os_client = self._initialize_os_client()
        self.files_client = self._initialize_files_client()

    def _initialize_sharepoint_client(
        self, config: SharepointConfig
    ) -> SharepointClient:
        """Initialize the HTTP client."""
        try:
            return SharepointClient(
                SharepointClientInitArgs(config, client_name="sharepoint_rest_api")
            )
        except ConnectionError as error:
            raise SharepointClientError(error)

    def _initialize_thies_ftp_client(self, config: FtpClientConfig) -> FTPClient:
        """Initialize the FTP client."""
        try:
            return FTPClient(FtpClientInitArgs(config, client_name="aioftp_client"))
        except RuntimeError as error:
            raise FtpClientError(error)

    def _initialize_os_client(self) -> DirectoryClient:
        return DirectoryClient(DirectoryClientArgs(client_name="os_client"))

    def _initialize_files_client(self) -> FilesClient:
        return FilesClient(FilesClientInitArgs(client_name="aiofiles_client"))

    async def _validate_sharepoint_current_folders(self):
        async with self.sharepoint_client:
            folder_base_path = "/".join(
                self.sharepoint_folders_path[0].split("/")[0:-1]
            )
            relative_url = f"{self.sharepoint_base_url}/{folder_base_path}"
            response = await self.sharepoint_client.list_folders(
                SpListFoldersArgs(relative_url)
            )

        current_folders = [item["Name"] for item in response["value"]]  # type: ignore

        for folder_path in self.sharepoint_folders_path:
            folder_name = folder_path.split("/")[-1]
            if folder_name not in current_folders:
                raise SharePointDirectoryError(
                    reason=f"The current folder '{folder_name}' doesn't exist."
                )

    async def fetch_cloud_file_names(self) -> set[str]:
        """Fetch file names from the RCER cloud."""
        await self._validate_sharepoint_current_folders()
        try:
            cloud_files = set()
            async with self.sharepoint_client:
                for folder_path in self.sharepoint_folders_path:
                    folder_name = folder_path.split("/")[-1]
                    relative_url = f"{self.sharepoint_base_url}/{folder_path}"
                    args = SpListFilesArgs(folder_relative_url=relative_url)
                    response = await self.sharepoint_client.list_files(args)
                    cloud_files.update(
                        {f"{folder_name}_{item['Name']}" for item in response["value"]}  # type: ignore
                    )
            return cloud_files
        except Exception as error:
            raise SharePointFetchingError(reason=error)

    async def fetch_thies_file_names(self) -> set[str]:
        """Fetch file names from the THIES FTP server."""
        try:
            thies_files = set()
            for folder_path in self.ftp_server_folders_path:
                # AV for average, and EXT for extreme.
                prefix = "AVG" if "AV" in folder_path else "EXT"
                files = await self.thies_ftp_client.list_files(
                    FtpListFilesArgs(path=folder_path)
                )
                files_names = {f"{prefix}_{name}" for name in files}
                thies_files.update(files_names)
            return thies_files
        except ConnectionRefusedError as error:
            raise ThiesConnectionError(reason=error)
        except ConnectionAbortedError as error:
            raise ThiesFetchingError(reason=error)

    async def fetch_thies_file_content(self) -> dict[str, bytes]:
        """Fetch the content of files from the THIES FTP server."""
        try:
            content_files = {}
            for file in self.uploading:
                prefix, filename = file.split("_", 1)
                # The first path is for AVG files. The second file is for EXT files
                folder_path = next(
                    (
                        path
                        for path in self.ftp_server_folders_path
                        if prefix == ("AVG" if "AV" in path else "EXT")
                    ),
                    self.ftp_server_folders_path[0],  # Default to the first path
                )
                file_path = f"{folder_path}/{filename}"
                content = await self.thies_ftp_client.read_file(
                    FtpReadFileArgs(file_path)
                )
                self.logger.debug(
                    "[thies_synchronization_lib] Fetching file '%s' from '%s'.",
                    file,
                    folder_path,
                )
                content_files[file] = (
                    content  # Save file content with its original name.
                )
            return content_files
        except ConnectionRefusedError as error:
            raise ThiesConnectionError(reason=error)
        except ConnectionAbortedError as error:
            raise ThiesFetchingError(reason=error)

    async def upload_thies_files_to_sharepoint(
        self, files: dict
    ) -> dict[str, list[str]]:
        """Upload files to SharePoint and categorize the results."""
        upload_results = {"failed_files": [], "new_files": []}

        async with self.sharepoint_client:
            for file, file_content in files.items():
                try:
                    origin, file_name = file.split("_", 1)
                    # Check if the first folder is for AVG, otherwise assume it's for EXT
                    if "AVG" in self.sharepoint_folders_path[0]:
                        avg_folder = self.sharepoint_folders_path[0]
                        ext_folder = self.sharepoint_folders_path[1]
                    else:
                        avg_folder = self.sharepoint_folders_path[1]
                        ext_folder = self.sharepoint_folders_path[0]

                    folder_path = avg_folder if origin == "AVG" else ext_folder

                    relative_url = f"{self.sharepoint_base_url}/{folder_path}"

                    args = SpUploadFileArgs(
                        folder_relative_url=relative_url,
                        file_content=file_content,
                        file_name=file_name,
                    )
                    await self.sharepoint_client.upload_file(args)
                    upload_results["new_files"].append(file)
                    self.logger.debug(
                        "[thies_synchronization_lib] File '%s' from '%s' uploaded successfully to '%s' ✅",
                        file_name,
                        folder_path,
                        relative_url,
                    )

                except ConnectionError as error:
                    self.logger.error(
                        "[thies_synchronization_lib] Unexpected error from with file '%s'",
                        file_name,
                    )
                    upload_results["failed_files"].append(
                        f"{file} (Error: {str(error)})"
                    )

        if upload_results["failed_files"]:
            raise SharePointUploadError(
                reason="Files failed to upload: "
                + ", ".join(upload_results["failed_files"])
            )

        return upload_results

    async def _sync_pending_files(self, thies_files: set, cloud_files: set) -> set:
        uploading = thies_files - cloud_files

        # Update content of the daily files
        daily_files = {
            prefix + datetime_to_str(today(), date_format="%Y%m%d") + ".BIN"
            for prefix in ["EXT_", "AVG_"]
        }
        for file in daily_files:
            if file in thies_files:
                uploading.add(file)

        return uploading

    async def _extract_thies_daily_statistics(self) -> None:
        # Create the folder thies-daily-files if doesnt exists
        self.logger.info("[thies_synchronization_lib] Creating Daily files directory")
        base_folder = "thies-daily-files"
        if not await self.os_client.isdir(base_folder):
            for dest_folder in {"ARCH_AV1", "ARCH_EX1"}:
                await self.os_client.makedirs(
                    self.os_client.join_paths(base_folder, dest_folder)
                )
        else:
            self.logger.info(
                "[thies_synchronization_lib] Thies daily files already exists"
            )

        # Read the daily files and save each data in the folder
        daily_files = [
            prefix + datetime_to_str(today(), date_format="%Y%m%d") + ".BIN"
            for prefix in ["AVG_", "EXT_"]
        ]
        # Receive from FTP server and  write the file in thies-daily-files
        for file in daily_files:
            prefix, filename = file.split("_", 1)
            # The first path is for AVG files. The second file is for EXT files
            folder_path = next(
                (
                    path
                    for path in self.ftp_server_folders_path
                    if prefix == ("AVG" if "AV" in path else "EXT")
                ),
                self.ftp_server_folders_path[0],  # Default to the first path
            )
            # Retrieve the AVG or EXT file
            file_path = f"{folder_path}/{filename}"
            try:
                content = await self.thies_ftp_client.read_file(
                    FtpReadFileArgs(file_path)
                )
            except FileNotFoundError as error:
                raise ThiesFetchingError(reason=str(error))
            # Destination local folder
            self.logger.debug(file_path)

            dest_folder = "ARCH_AV1" if prefix == "AVG" else "ARCH_EX1"
            await self.files_client.write(
                WriteArgs(
                    file_name=filename,
                    file_content=content,
                    mode="wb",
                    destination_path=f"{base_folder}/{dest_folder}",
                )
            )
            # Retrieve the DESCFILE and save if is not in the base folder
            descfile_name = "DESCFILE.INI"
            if not await self.os_client.path_exists(
                self.os_client.join_paths(base_folder, dest_folder, descfile_name)
            ):
                descfile_path = f"{folder_path}/{descfile_name}"
                descfile_content = await self.thies_ftp_client.read_file(
                    FtpReadFileArgs(descfile_path)
                )
                await self.files_client.write(
                    WriteArgs(
                        file_name=descfile_name,
                        file_content=descfile_content,
                        mode="wb",
                        destination_path=f"{base_folder}/{dest_folder}",
                    )
                )
            else:
                self.logger.debug(
                    "[thies_synchronization_lib] DESCFILE.INI already exists in %s",
                    dest_folder,
                )
        # Read the files with ThiesDayData class
        await create_thies_daily_statistics_file(
            self.os_client, self.logger, daily_files
        )

    async def execute(self):
        """Synchronize data from the THIES Center to the cloud."""
        self.logger.debug("[thies_synchronization_lib] Starting ...")
        try:
            thies_files = await self.fetch_thies_file_names()
        except RuntimeError as error:
            raise FtpClientError(error)
        self.logger.debug(
            "[thies_synchronization_lib] Total files fetched from THIES: %s",
            str(len(thies_files)),
        )
        try:
            cloud_files = await self.fetch_cloud_file_names()
        except RuntimeError as error:
            raise SharepointClient(error)  # type: ignore
        self.logger.debug(
            "[thies_synchronization_lib] Total files fetched from Sharepoint: %s",
            str(len(cloud_files)),
        )
        self.uploading = await self._sync_pending_files(thies_files, cloud_files)
        # Extract thies statistics for SAVIIA Sensors
        await self._extract_thies_daily_statistics()

        if not self.uploading:
            raise EmptyDataError(reason="No files to upload.")
        # Fetch the content of the files to be uploaded from THIES FTP Server
        thies_fetched_files = await self.fetch_thies_file_content()

        # Upload the fetched files to SharePoint and gather statistics
        upload_statistics = await self.upload_thies_files_to_sharepoint(
            thies_fetched_files
        )
        self.logger.info(upload_statistics)
        self.logger.debug(
            "[thies_synchronization_lib] All the files were uploaded successfully 🎉"
        )

        return parse_execute_response(thies_fetched_files, upload_statistics)  # type: ignore
